use std::{
    any::Any,
    sync::atomic::{AtomicU64, Ordering},
};

use crate::backend::{Backend, Combine, Render, RenderResult, TargetBackend};

#[doc(hidden)]
pub mod __internal;
use __internal::Node;

pub struct View<B: Backend = TargetBackend> {
    id: Id,
    node: Box<dyn Node<B>>,
}

pub trait Construct: Sized {
    type Params: Default;

    fn construct(params: Self::Params) -> Self;
}

pub trait ConstructGroup<B: Backend>: Construct {
    fn construct_group<I>(params: Self::Params, children: I) -> Self
    where
        I: Iterator<Item = View<B>>;
}

impl<B: Backend> View<B> {
    pub fn new<N: Node<B> + 'static>(node: N) -> Self {
        Self {
            id: Id::new(),
            node: Box::new(node),
        }
    }

    pub fn id(&self) -> Id {
        self.id
    }
}

impl<B: Backend> Render<B> for View<B> {
    fn render(&self, context: &mut B::Context) -> RenderResult<B> {
        self.node.render(context)
    }
}

pub trait NativeComponent<B: Backend = TargetBackend>: Any + Send + Render<B> {}

pub struct Template<B: Backend = TargetBackend> {
    content: Vec<View<B>>,
}

impl<B: Backend> FromIterator<View<B>> for Template<B> {
    fn from_iter<T: IntoIterator<Item = View<B>>>(iter: T) -> Self {
        Self {
            content: iter.into_iter().collect(),
        }
    }
}

impl<B: Backend> Render<B> for Template<B>
where
    View<B>: Render<B>,
{
    fn render(&self, context: &mut B::Context) -> RenderResult<B> {
        let mut result: Option<B::Output> = None;
        for view in &self.content {
            if let Some(x) = view.render(context)? {
                result = Some(match result {
                    Some(result) => result.combine(x),
                    None => x,
                })
            }
        }
        Ok(result)
    }
}

pub trait GetTemplate<B: Backend = TargetBackend> {
    fn get_template(&self) -> Template<B>;
}

/// Globally unique ID for any instance of [`View`].
/// No ID will be issued twice for the entire lifetime of the application,
/// even IDs of views that have been destroyed.  You shouldn't make any
/// assumptions about this type beyond that.
#[derive(Copy, Clone, Debug, Hash, Eq, PartialEq)]
pub struct Id(u64);

static ID_COUNTER: AtomicU64 = AtomicU64::new(Id::SENTINEL);

impl Id {
    const SENTINEL: u64 = 0;

    #[allow(clippy::new_without_default)]
    pub fn new() -> Self {
        let id = ID_COUNTER
            .fetch_add(1, Ordering::Relaxed)
            .checked_add(1)
            .expect("ID counter should never overflow");

        Self(id)
    }

    pub(crate) const fn sentinel() -> Self {
        Self(Self::SENTINEL)
    }

    pub(crate) const fn is_sentinel(&self) -> bool {
        self.0 == Self::SENTINEL
    }
}
